/*----------------------------------------------------------------------
  File    : lvq.h
  Contents: learning vector quantization with some extensions
  Author  : Christian Borgelt
  History : 2003.02.17 file created from file cluster.h
            2003.03.11 weight parameter added to function lvq_reg
            2003.03.15 organization of quantization prototypes changed
            2003.05.15 activation normalization modes added
            2003.06.07 function lvq_scale added
            2003.06.08 parameter radius added to function lvq_type
            2003.08.11 adapted to new module attmap
            2003.08.12 adapted to new module nstats
            2003.08.15 adapted to new module radfn
            2007.01.10 use of attribute maps changed
            2013.08.19 adapted to definition of type DIMID in matrix.h
            2018.09.10 extended functions (attset, attmap) removed
----------------------------------------------------------------------*/
#ifndef __LVQ__
#define __LVQ__
#include <stdio.h>
#include <math.h>
#include "fntypes.h"
#include "radfn.h"
#ifdef LVQ_PARSE
#ifndef NST_PARSE
#define NST_PARSE
#endif
#ifndef SCN_SCAN
#define SCN_SCAN
#endif
#endif
#include "nstats.h"
#include "matrix.h"

/*----------------------------------------------------------------------
  Preprocessor Definitions
----------------------------------------------------------------------*/
/* --- prototype flags --- */
#define LVQ_CENTER       0x00   /* use prototype center (mandatory) */
#define LVQ_SIZE         0x01   /* use prototype size (radius) */

/* --- initialization modes --- */
/*      LVQ_CENTER          0      center of the data space */
#define LVQ_UNIFORM         1   /* samples from uniform distribution */
#define LVQ_DIAG            2   /* diagonal of data space */
#define LVQ_LATIN           3   /* latin hypercube sampling */
#define LVQ_POINTS          4   /* given points in data space */

/* --- normalization modes --- */
#define LVQ_NONE            0   /* no normalization */
#define LVQ_SUM1            1   /* normalize to sum 1 */
#define LVQ_MAX1            2   /* normalize to maximum 1 */
#define LVQ_HARD            3   /* hard assignment/winner takes all */

/* --- update methods --- */
#define LVQ_DIST            0   /* average distance to center */
#define LVQ_SQUARE          1   /* average squared distance */
#define LVQ_VAR             2   /* isotropic variance */
#define LVQ_WEIGHT          3   /* weight of assigned patterns */
#define LVQ_REPLACE    0x8000   /* replace old radius with new */

/* --- description modes --- */
#define LVQ_TITLE      0x0001   /* print a title (as a comment) */
#define LVQ_INFO       0x0002   /* print add. info. (as a comment) */
#define LVQ_RBFNET     0x0004   /* describe as part of RBF network */

/*----------------------------------------------------------------------
  Type Definitions
----------------------------------------------------------------------*/
typedef struct {                /* --- quantization prototype --- */
  double *ctr;                  /* center (reference vector) */
  double *dif;                  /* difference vector to center */
  double *chg;                  /* change of center coordinates */
  double var;                   /* variance (squared radius) */
  double wgt;                   /* (relative) prototype weight */
  double d2;                    /* squared distance of pattern */
  double act;                   /* neuron activation */
  double sqr;                   /* weighted sum of squared distances */
  double sum;                   /* weighted sum of distances */
  double spw;                   /* sum of (assigned) pattern weights */
  double buf;                   /* buffer for variance/radius update */
} QPROT;                        /* (quantization prototype) */

typedef struct {                /* --- LVQ network --- */
  DIMID  dim;                   /* number of dimensions */
  DIMID  cnt;                   /* number of neurons */
  DIMID  init;                  /* number of initialized neurons */
  int    type;                  /* current type of prototypes */
  int    mode;                  /* normalization mode */
  int    method;                /* parameter update method */
  int    steps;                 /* number of update steps */
  RADFN  *actfn;                /* activation function */
  double params[3];             /* parameters of the act. function */
  double wtarf;                 /* winner takes all radius factor */
  double lrate;                 /* initial learning rate */
  double lrcur;                 /* current learning rate */
  double decay;                 /* learning rate decay parameter */
  double exp;                   /* exponent for neuron activation */
  double szscl;                 /* prototype size scaling factor */
  double reg;                   /* sum of (regist.) pattern weights */
  double spw;                   /* sum of (processed) pattern weights */
  NSTATS *nst;                  /* simple numerical statistics */
  double *vec;                  /* buffer for a data vector */
  double *buf;                  /* buffer for init./aggregation */
  QPROT  *qps;                  /* quantization prototypes */
} LVQNET;                       /* (LVQ network) */

/*----------------------------------------------------------------------
  Functions
----------------------------------------------------------------------*/
extern LVQNET* lvq_create  (DIMID dim, DIMID cnt);
extern void    lvq_delete  (LVQNET *lvq);
extern DIMID   lvq_dim     (LVQNET *lvq);
extern DIMID   lvq_inpcnt  (LVQNET *lvq);
extern DIMID   lvq_cnt     (LVQNET *lvq);
extern DIMID   lvq_outcnt  (LVQNET *lvq);
extern DIMID   lvq_veccnt  (LVQNET *lvq);

extern NSTATS* lvq_nstats  (LVQNET *lvq);
extern void    lvq_reg     (LVQNET *lvq, const double *vec,
                            double weight);
extern void    lvq_value   (LVQNET *lvq, DIMID index, double value);

extern void    lvq_type    (LVQNET *lvq, int type, double radius);
extern void    lvq_actfn   (LVQNET *lvq, RADFN actfn, double *params);
extern void    lvq_norm    (LVQNET *lvq, int mode, double wtarf);

extern void    lvq_init    (LVQNET *lvq, int mode, double range,
                            RANDFN randfn, double *vec);
extern void    lvq_exp     (LVQNET *lvq, double exp);
extern void    lvq_scale   (LVQNET *lvq, double scale);
extern void    lvq_method  (LVQNET *lvq, int method);
extern void    lvq_lrate   (LVQNET *lvq, double lrate, double decay);

extern int     lvq_setup   (LVQNET *lvq);
extern double  lvq_sqrdist (LVQNET *lvq, DIMID pid, const double *vec);
extern double  lvq_dist    (LVQNET *lvq, DIMID pid, const double *vec);
extern DIMID   lvq_exec    (LVQNET *lvq, double *vec, double *acts);
extern DIMID   lvq_aggr    (LVQNET *lvq, double *vec, double weight);
extern double  lvq_update  (LVQNET *lvq);
extern void    lvq_sort    (LVQNET *lvq);

extern double* lvq_center  (LVQNET *lvq, DIMID pid);
extern double  lvq_var     (LVQNET *lvq, DIMID pid);
extern double  lvq_activ   (LVQNET *lvq, DIMID pid);
extern void    lvq_unscale (LVQNET *lvq);

extern int     lvq_desc    (LVQNET *lvq, FILE *file,
                            int mode, int maxlen);
#ifdef LVQ_PARSE
extern LVQNET* lvq_parse   (SCANNER *scan, DIMID dim, int hdr);
#endif

/*----------------------------------------------------------------------
  Preprocessor Definitions
----------------------------------------------------------------------*/
#define lvq_dim(n)         ((n)->dim)
#define lvq_inpcnt(n)      ((n)->dim)
#define lvq_cnt(n)         ((n)->cnt)
#define lvq_outcnt(n)      ((n)->cnt)
#define lvq_veccnt(n)      ((n)->cnt)

#define lvq_nstats(n)      ((n)->nst)
#define lvq_reg(n,v,w)     nst_reg((n)->nst, v, w)

#define lvq_dist(n,i,v)    sqrt(lvq_sqrdist(n,i,v))

#define lvq_norm(n,m,f)    ((n)->mode   = (m), (n)->wtarf = (f))
#define lvq_exp(n,e)       ((n)->exp    = (e))
#define lvq_scale(n,s)     ((n)->szscl  = (s))
#define lvq_method(n,m)    ((n)->method = (m))
#define lvq_lrate(n,r,d)   ((n)->lrcur  = fabs((n)->lrate = (r)), \
                            (n)->decay  = (d))

#define lvq_center(n,i)    ((n)->qps[i].ctr)
#define lvq_var(n,i)       ((n)->qps[i].var)
#define lvq_activ(n,i)     ((n)->qps[i].act)

#endif
